from inkletter.codeblock import codeblock_from_string, CodeBlockResolver


def resolve_block(block, indent_size=2):
    resolver = CodeBlockResolver(indent_size=indent_size)
    return resolver.resolve(block)


def test_codeblock_simple_code():
    content = """\
def foo():
  return 42"""

    block = codeblock_from_string(content)
    result = resolve_block(block)

    assert result == content


def test_codeblock_multiple_indent_levels():
    content = """\
def outer():
  def inner():
    return 42"""

    block = codeblock_from_string(content)
    result = resolve_block(block)

    assert result == content


def test_codeblock_empty_lines():
    content = """\
def foo():

  return 42"""

    block = codeblock_from_string(content)
    result = resolve_block(block)

    assert result == content


def test_codeblock_only_empty_lines():
    content = """

"""

    block = codeblock_from_string(content)
    result = resolve_block(block)

    assert result == ""


def test_codeblock_no_indentation():
    content = """\
print('Hello')
print('World')"""

    block = codeblock_from_string(content)
    result = resolve_block(block)

    assert result == content


def test_codeblock_mixed_indent_and_unindented_lines():
    content = """\
print('Start')
  print('Indented')
print('End')"""

    block = codeblock_from_string(content)
    result = resolve_block(block)

    assert result == content


def test_codeblock_custom_indent_size():
    content = """\
def foo():
    print('Hello')
    print('World')"""

    block = codeblock_from_string(content, indent_size=4)
    result = resolve_block(block, indent_size=4)

    assert result == content



def parse_markdown(markdown_text: str) -> list:
    """Parse le Markdown enrichi et retourne l'AST."""


def ast_to_mjml(ast: list) -> str:
    """Transforme l'AST en MJML."""


def markdown_to_mjml(markdown_text: str) -> str:
    """Pipeline complet: Markdown -> AST -> MJML."""



from contextlib import contextmanager
import html

from inkletter.ast import *
from inkletter.codeblock import (
    CodeBlock,
    CodeBlockResolver,
    TextElement,
)
from inkletter.visitors.generic import NodeVisitor


class Codegen(NodeVisitor):
    def __init__(self, use_style=False):
        super().__init__()
        self.root = CodeBlock()
        self.resolver = CodeBlockResolver()
        self.current = self.root
        self.use_style = use_style

    def get_code(self):
        return self.resolver.resolve(self.root)

    @contextmanager
    def block_tag(self, name, attrs=None, self_closing=False, inline=False):
        attrs = attrs or {}
        attrs_str = "".join(f' {k}="{v}"' for k, v in attrs.items())
        if self_closing:
            self.current.add_text(f"<{name}{attrs_str}/>")
            if not inline:
                self.current.add_newline()
            yield
            return

        parent = self.current

        block = CodeBlock()
        if (
            not inline
            and parent.elements
            and isinstance(parent.elements[-1], TextElement)
        ):
            block.add_newline()

        block.add_text(f"<{name}{attrs_str}>")

        if not inline:
            block.add_newline()
            block.add_indent()

        parent.add_codeblock(block)
        self.current = block

        yield

        self.current = parent
        if not inline:
            last_element = parent.elements[-1]
            assert last_element == block
            if isinstance(last_element.elements[-1], TextElement):
                last_element.add_newline()
            self.current.add_dedent()

        self.current.add_text(f"</{name}>")

        if not inline:
            self.current.add_newline()

    @contextmanager
    def ensure_open_column(self, node):
        if node.annotations.get("requires_column"):
            with self.block_tag("mj-section"):
                with self.block_tag("mj-column"):
                    yield
        else:
            yield

    @contextmanager
    def ensure_open_text(self, node, inline=False):
        if node.annotations.get("requires_text"):
            with self.block_tag("mj-text", inline=inline):
                yield
        else:
            yield

    def visit_Document(self, node, scope):
        with self.block_tag("mjml"):
            if self.use_style:
                with self.block_tag("mj-head"):
                    with self.block_tag("mj-attributes"):
                        with self.block_tag(
                            "mj-section",
                            attrs={"padding": "0"},
                            self_closing=True,
                        ):
                            pass
            with self.block_tag("mj-body"):
                self.generic_visit(node, scope)

    def visit_Paragraph(self, node, scope):
        with self.ensure_open_column(node):
            self.generic_visit(node, scope)

    def visit_Heading(self, node, scope):
        with self.ensure_open_column(node):
            with self.ensure_open_text(node):
                with self.block_tag(f"h{node.level}", inline=True):
                    self.generic_visit(node, scope)

    def visit_BlockText(self, node, scope):
        with self.ensure_open_column(node):
            with self.ensure_open_text(node):
                self.generic_visit(node, scope)

    def visit_BlockQuote(self, node, scope):
        with self.ensure_open_column(node):
            with self.block_tag(
                "mj-text", attrs={"font-style": "italic", "color": "#555555"}
            ):
                self.generic_visit(node, scope)

    def visit_BlockCode(self, node, scope):
        with self.ensure_open_column(node):
            with self.ensure_open_text(node):
                with self.block_tag("pre"):
                    self.current.add_text(html.escape(node.code), indented=False)

    def visit_ThematicBreak(self, node, scope):
        with self.ensure_open_column(node):
            with self.block_tag(
                "mj-divider",
                attrs={"border-color": "#cccccc", "border-width": "1px"},
                self_closing=True,
            ):
                pass

    def visit_LineBreak(self, node, scope):
        self.current.add_text("<br/>")
        self.current.add_newline()

    def visit_SoftBreak(self, node, scope):
        self.current.add_text("<br/>")
        self.current.add_newline()

    def visit_BlankLine(self, node, scope):
        pass

    def visit_Image(self, node, scope):
        with self.ensure_open_column(node):
            attrs = {"src": node.url}
            if node.alt_text:
                attrs["alt"] = node.alt_text.value
            if node.title:
                attrs["title"] = node.title

            if node.annotations.get("requires_manual_image"):
                tag = "img"
                attrs["style"] = "max-width: 100%; height: auto;"
            else:
                tag = "mj-image"

            with self.block_tag(tag, attrs=attrs, self_closing=True):
                pass

    def visit_ImageLink(self, node, scope):
        with self.ensure_open_column(node):
            attrs = {"src": node.img.url, "href": node.href}
            if node.img.alt_text:
                attrs["alt"] = node.img.alt_text.value
            if node.img.title:
                attrs["title"] = node.img.title

            if node.img.annotations.get("requires_manual_image"):
                tag = "img"
                attrs["style"] = "max-width: 100%; height: auto;"
            else:
                tag = "mj-image"

            with self.block_tag(tag, attrs=attrs, self_closing=True, inline=True):
                pass

    def visit_List(self, node, scope):
        with self.ensure_open_column(node):
            with self.ensure_open_text(node):
                tag = "ol" if node.ordered else "ul"
                attrs = (
                    {"style": "list-style-type: none;"}
                    if node.annotations.get("is_task_list")
                    else {}
                )
                with self.block_tag(tag, attrs=attrs):
                    self.generic_visit(node, scope)

    def visit_ListItem(self, node, scope):
        with self.block_tag("li"):
            self.generic_visit(node, scope)

    def visit_TaskListItem(self, node, scope):
        checkbox = "☑" if node.checked else "☐"
        with self.block_tag("li"):
            block = node.children[0]
            assert isinstance(block, BlockText)
            self.current.add_text(checkbox + " ")
            for child in block.get_children():
                self.visit(child, scope)
            self.current.add_newline()
            for extra_child in node.children[1:]:
                self.visit(extra_child, scope)

    def visit_Table(self, node, scope):
        with self.ensure_open_column(node):
            tag = (
                "table"
                if node.annotations.get("requires_manual_table", False)
                else "mj-table"
            )
            with self.block_tag(tag):
                if node.header:
                    self.visit(node.header, scope)
                for row in node.rows:
                    self.visit(row, scope)

    def visit_TableHeader(self, node, scope):
        with self.block_tag("tr"):
            for head in node.headers:
                self.visit(head, scope)

    def visit_TableRow(self, node, scope):
        with self.block_tag("tr"):
            for cell in node.row:
                self.visit(cell, scope)

    def visit_TableHeaderCell(self, node, scope):
        attrs = {"align": node.align} if node.align else {}
        with self.block_tag("th", attrs=attrs, inline=True):
            self.generic_visit(node, scope)
        self.current.add_newline()

    def visit_TableCell(self, node, scope):
        attrs = {"align": node.align} if node.align else {}
        with self.block_tag("td", attrs=attrs, inline=True):
            self.generic_visit(node, scope)
        self.current.add_newline()

    def visit_LiteralText(self, node, scope):
        with self.ensure_open_text(node):
            self.current.add_text(node.value)

    def visit_Emphasis(self, node, scope):
        with self.ensure_open_text(node):
            with self.block_tag("em", inline=True):
                self.generic_visit(node, scope)

    def visit_Strong(self, node, scope):
        with self.ensure_open_text(node):
            with self.block_tag("strong", inline=True):
                self.generic_visit(node, scope)

    def visit_StrikeThrough(self, node, scope):
        with self.ensure_open_text(node):
            with self.block_tag("del", inline=True):
                self.generic_visit(node, scope)

    def visit_Link(self, node, scope):
        with self.ensure_open_text(node):
            attrs = {"href": node.href}
            if node.title:
                attrs["title"] = node.title
            with self.block_tag("a", attrs=attrs, inline=True):
                self.generic_visit(node, scope)

    def visit_CodeSpan(self, node, scope):
        with self.ensure_open_text(node):
            with self.block_tag("code", inline=True):
                self.current.add_text(html.escape(node.code))



import mistune

from inkletter.ast import *
from inkletter.md_to_ast import ASTRenderer
from inkletter.scope import ScopeStack
from inkletter.visitors.annotation import Annotation
from inkletter.visitors.merger import BlockTextMerger
from inkletter.visitors.tree import print_tree


def generate_ast(markdown_input):
    renderer = ASTRenderer()
    markdown = mistune.create_markdown(
        renderer=renderer,
        plugins=[
            "table",
            "strikethrough",
            "task_lists",
        ],
    )
    ast = markdown(markdown_input)
    Annotation().visit(ast, scope=ScopeStack())
    print("AST:")
    print_tree(ast)
    return ast


def merge_ast(ast):
    splitter = BlockTextMerger()
    splitter.visit(ast, scope=None)
    print("AST after splitting:")
    print_tree(ast)
    return ast


def test_merge_text():
    markdown_input = """\
This is a text
This is anoter text
"""

    # === AST BEFORE ===
    ast = generate_ast(markdown_input)
    assert isinstance(ast, Document)
    assert len(ast.children) == 1
    paragraph = ast.children[0]
    assert isinstance(paragraph, Paragraph)
    assert len(paragraph.children) == 3  # LiteralText, SoftBreak, LiteralText
    assert isinstance(paragraph.children[0], LiteralText)
    assert paragraph.children[0].value == "This is a text"
    assert isinstance(paragraph.children[1], SoftBreak)
    assert isinstance(paragraph.children[2], LiteralText)
    assert paragraph.children[2].value == "This is anoter text"

    # === AST AFTER MERGE ===
    merged_ast = merge_ast(ast)
    assert isinstance(ast, Document)
    assert len(ast.children) == 1
    paragraph = merged_ast.children[0]
    assert isinstance(paragraph, Paragraph)
    assert len(paragraph.children) == 1
    block = paragraph.children[0]
    assert isinstance(block, BlockText)
    assert len(block.children) == 3
    assert isinstance(block.children[0], LiteralText)
    assert block.children[0].value == "This is a text"
    assert isinstance(block.children[1], SoftBreak)
    assert isinstance(block.children[2], LiteralText)
    assert block.children[2].value == "This is anoter text"



from inkletter.ast import *
from inkletter.visitors.generic import NodeVisitor


class Annotation(NodeVisitor):

    def mark_column_if_needed(self, node, scope):
        if not scope.get("in_column", False):
            node.annotations["requires_column"] = True
            scope.set("in_column", True)

    def mark_text_if_needed(self, node, scope):
        if (
            not scope.get("in_text", False)
            and not scope.get("is_in_table_cell", False)
            and not scope.get("is_in_list_item", False)
        ):
            node.annotations["requires_text"] = True
            scope.set("in_text", True)

    def mark_manual_table_if_needed(self, node, scope):
        is_in_list = scope.get("is_in_list", False)
        if is_in_list:
            node.annotations["requires_manual_table"] = True

    def mark_manual_image_if_needed(self, node, scope):
        is_in_table = scope.get("is_in_table", False)
        is_in_heading = scope.get("is_in_heading", False)
        is_in_blockquote = scope.get("is_in_blockquote", False)
        is_in_list = scope.get("is_in_list", False)
        if is_in_table or is_in_heading or is_in_blockquote or is_in_list:
            node.annotations["requires_manual_image"] = True

    def mark_column_and_text_if_needed(self, node, scope):
        self.mark_column_if_needed(node, scope)
        self.mark_text_if_needed(node, scope)

    def visit_List(self, node, scope):
        scope.push(node)
        scope.set("is_in_list", True)
        self.mark_column_and_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        node.annotations["is_task_list"] = scope.get("has_task_item", False)
        scope.pop(node)

    def visit_TaskListItem(self, node, scope):
        assert isinstance(scope.stack[-1]["node"], List)
        scope.set("has_task_item", True)
        scope.push(node)
        scope.set("is_in_list_item", True)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_ListItem(self, node, scope):
        assert isinstance(scope.stack[-1]["node"], List)
        scope.set("has_task_item", False)
        scope.push(node)
        scope.set("is_in_list_item", True)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Paragraph(self, node, scope):
        scope.push(node)
        self.mark_column_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_BlockText(self, node, scope):
        scope.push(node)
        self.mark_column_and_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_BlockQuote(self, node, scope):
        scope.push(node)
        scope.set("is_in_blockquote", True)
        self.mark_column_and_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_BlockCode(self, node, scope):
        scope.push(node)
        self.mark_column_and_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Table(self, node, scope):
        scope.push(node)
        scope.set("is_in_table", True)
        self.mark_column_if_needed(node, scope)
        self.mark_manual_table_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_TableHeaderCell(self, node, scope):
        scope.push(node)
        scope.set("is_in_table_cell", True)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_TableCell(self, node, scope):
        scope.push(node)
        scope.set("is_in_table_cell", True)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Image(self, node, scope):
        scope.push(node)
        self.mark_column_if_needed(node, scope)
        self.mark_manual_image_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_ImageLink(self, node, scope):
        scope.push(node)
        self.mark_column_if_needed(node, scope)
        self.mark_manual_image_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Heading(self, node, scope):
        scope.push(node)
        scope.set("is_in_heading", True)
        self.mark_column_and_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_LiteralText(self, node, scope):
        scope.push(node)
        self.mark_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Text(self, node, scope):
        scope.push(node)
        self.mark_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Emphasis(self, node, scope):
        scope.push(node)
        self.mark_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_Strong(self, node, scope):
        scope.push(node)
        self.mark_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_StrikeThrough(self, node, scope):
        scope.push(node)
        self.mark_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_CodeSpan(self, node, scope):
        scope.push(node)
        self.mark_text_if_needed(node, scope)
        self.generic_visit(node, scope)
        scope.pop(node)

    def visit_ThematicBreak(self, node, scope):
        scope.push(node)
        self.mark_column_if_needed(node, scope)
        scope.pop(node)

    def visit_LineBreak(self, node, scope):
        scope.push(node)
        self.mark_column_and_text_if_needed(node, scope)
        scope.pop(node)

    def visit_SoftBreak(self, node, scope):
        scope.push(node)
        self.mark_column_and_text_if_needed(node, scope)
        scope.pop(node)



class NodeVisitor:
    def visit(self, node, scope):
        """Dispatch to the appropriate visit method or fallback to generic_visit."""
        method_name = f"visit_{node.__class__.__name__}"
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node, scope)

    def generic_visit(self, node, scope):
        """Default visit method if no visit_XXX found."""
        for child in node.get_children():
            self.visit(child, scope)



from pathlib import Path
import tempfile
import webbrowser

import click
from jinja2 import Environment, FileSystemLoader

from inkletter.md_to_html import parse_markdown_to_html, parse_mjml_to_html
from inkletter.md_to_mjml import parse_markdown_to_mjml

BASE_DIR = Path(__file__).resolve().parent
TEMPLATES_DIR = BASE_DIR / "templates"

@click.group()
def cli():
    """Inkletter CLI — Convert and preview Markdown as MJML"""
    pass


@cli.command()
@click.argument("filepath", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option("-o", "--output", type=click.Path(dir_okay=False, writable=True, path_type=Path), help="Output HTML file path")
def preview(filepath: Path, output: Path | None):
    """
    Convert a Markdown file into MJML, then into HTML, and preview it in a browser.
    """
    try:
        # Load Markdown
        markdown_text = filepath.read_text(encoding="utf-8")

        # Markdown → MJML
        mjml_code = parse_markdown_to_mjml(
            markdown_text,
            print_ast=False,
            use_style=True,
        )

        # MJML → HTML
        html_output = parse_mjml_to_html(mjml_code)

        # Render with Jinja2
        env = Environment(loader=FileSystemLoader(str(TEMPLATES_DIR)))
        template = env.get_template("preview.html")
        rendered = template.render(
            MD_CONTENT=markdown_text,
            MJML_CONTENT=mjml_code,
            HTML_CONTENT=html_output,
        )

        if output:
            output.write_text(rendered, encoding="utf-8")
            out_path = output
        else:
            with tempfile.NamedTemporaryFile("w", suffix=".html", encoding="utf-8", delete=False) as f:
                f.write(rendered)
                out_path = Path(f.name)

        webbrowser.open(f"file://{out_path.resolve().as_uri()}")

        click.secho(f"✅ Preview opened in browser:\n{out_path}", fg="green")

    except Exception as e:
        click.secho(f"❌ Error: {e}", fg="red", err=True)

@cli.command()
@click.argument("filepath", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option("-o", "--output", type=click.Path(dir_okay=False, writable=True, path_type=Path), help="Output HTML file path")
@click.option("--view/--no-view", default=False, help="Open the result in a browser (default: false)")
def md2mjml(filepath: Path, output: Path | None, view: bool):
    """Convert Markdown to MJML then HTML, optionally save and open it."""
    try:
        # Lire Markdown
        markdown_text = filepath.read_text(encoding="utf-8")

        # Markdown → HTML
        html_output = parse_markdown_to_html(markdown_text)


        if output:
            output.write_text(html_output, encoding="utf-8")
            out_path = output
        else:
            with tempfile.NamedTemporaryFile("w", suffix=".html", encoding="utf-8", delete=False) as f:
                f.write(html_output)
                out_path = Path(f.name)

        click.secho(f"✅ HTML written to: {out_path}", fg="green")

        if view:
            webbrowser.open(f"file://{out_path.resolve().as_uri()}")

    except Exception as e:
        click.secho(f"❌ Error: {e}", fg="red", err=True)




INDENT_SIZE = 2  # Number of spaces per indent level
NEWLINE = "\n"  # Newline character


# Base class for all elements
class CodeElement:
    def accept(self, visitor):
        raise NotImplementedError()


# Concrete element classes
class Indent(CodeElement):
    def accept(self, visitor):
        return visitor.visit_indent(self)


class Dedent(CodeElement):
    def accept(self, visitor):
        return visitor.visit_dedent(self)


class Newline(CodeElement):
    def accept(self, visitor):
        return visitor.visit_newline(self)


class TextElement(CodeElement):
    def __init__(self, text, indented=True):
        self.text = text
        self.indented = indented

    def accept(self, visitor):
        return visitor.visit_text(self)


# The main CodeBlock class
class CodeBlock(CodeElement):
    def __init__(self):
        self.elements = []

    def add_indent(self):
        self.elements.append(Indent())

    def add_dedent(self):
        self.elements.append(Dedent())

    def add_text(self, text, indented=True):
        self.elements.append(TextElement(text, indented=indented))

    def add_newline(self):
        self.elements.append(Newline())

    def add_codeblock(self, cb):
        self.elements.append(cb)

    def accept(self, visitor):
        return visitor.visit_codeblock(self)


# Visitor to resolve the CodeBlock to a string
class CodeBlockResolver:
    def __init__(self, indent_size=INDENT_SIZE, starting_indent=0):
        self.indent_size = indent_size
        self.starting_indent = starting_indent
        self.current_indent = starting_indent
        self.lines = []
        self.current_line = None

    def visit_indent(self, node):
        self.current_indent += 1

    def visit_dedent(self, node):
        if self.current_indent == 0:
            raise ValueError("Cannot dedent below zero")
        self.current_indent -= 1

    def visit_newline(self, node):
        self.lines.append(self.current_line or "")
        self.current_line = None

    def visit_text(self, node):
        if self.current_line is None:
            self.current_line = " " * (self.current_indent * self.indent_size) if node.indented else ""
        self.current_line += node.text

    def visit_codeblock(self, node):
        for element in node.elements:
            element.accept(self)

    def resolve(self, node):
        self.current_indent = self.starting_indent
        self.lines = []
        self.current_line = None
        node.accept(self)
        if self.current_line is not None:
            self.lines.append(self.current_line)
        return NEWLINE.join(l.rstrip() for l in self.lines)


def codeblock_from_string(content: str, indent_size: int = 2) -> CodeBlock:
    block = CodeBlock()
    lines = content.strip("\n").splitlines()

    current_indent_level = 0

    for line in lines:
        raw_line = line.rstrip()
        leading_spaces = len(line) - len(line.lstrip())
        line_indent_level = leading_spaces // indent_size

        # Adjust indent/dedent
        while current_indent_level < line_indent_level:
            block.add_indent()
            current_indent_level += 1
        while current_indent_level > line_indent_level:
            block.add_dedent()
            current_indent_level -= 1

        if raw_line.strip() == "":
            block.add_newline()
        else:
            block.add_text(raw_line.lstrip())
            block.add_newline()

    # Close any remaining indentation
    while current_indent_level > 0:
        block.add_dedent()
        current_indent_level -= 1

    return block



from inkletter.codeblock import CodeBlock, codeblock_from_string, CodeBlockResolver
from inkletter.md_to_ast import parse_markdown_to_ast
from inkletter.visitors.codegen import Codegen
from inkletter.visitors.tree import print_tree


def parse_markdown_to_mjml(markdown_text, print_ast=True, use_style=False):
    ast = parse_markdown_to_ast(markdown_text)
    if print_ast:
        print_tree(ast)
    codegen = Codegen(use_style=use_style)
    codegen.visit(ast, scope=None)
    code = codegen.get_code()
    return code


def wrap_mjml_body(content):
    if isinstance(content, str):
        content = codeblock_from_string(content)

    root = CodeBlock()
    root.add_text("<mjml>")
    root.add_newline()
    root.add_indent()

    root.add_text("<mj-body>")
    root.add_newline()
    root.add_indent()

    root.add_text("<mj-section>")
    root.add_newline()
    root.add_indent()

    root.add_text("<mj-column>")
    root.add_newline()
    root.add_indent()

    root.add_codeblock(content)

    root.add_dedent()
    root.add_text("</mj-column>")
    root.add_newline()

    root.add_dedent()
    root.add_text("</mj-section>")
    root.add_newline()

    root.add_dedent()
    root.add_text("</mj-body>")
    root.add_newline()

    root.add_dedent()
    root.add_text("</mjml>")
    root.add_newline()

    resolver = CodeBlockResolver()
    return resolver.resolve(root)



from inkletter.ast import *
from inkletter.visitors.generic import NodeVisitor


class BlockTextMerger(NodeVisitor):
    def merge_inline_sequences(self, children):
        new_children = []
        buffer = []

        def clean(_buf):
            while _buf:
                if isinstance(_buf[0], TextTerminal):
                    _buf.pop(0)
                    continue
                if isinstance(_buf[-1], TextTerminal):
                    buffer.pop(-1)
                    continue
                break

        def flush(_buf, _children):
            clean(_buf)
            if _buf:
                _children.append(BlockText(buffer.copy()))
                _buf.clear()

        for child in children:
            if isinstance(child, Text):
                buffer.append(child)
            elif isinstance(child, BlockText):
                flush(buffer, new_children)
                new_children.append(child)
            else:
                flush(buffer, new_children)
                new_children.append(child)

        flush(buffer, new_children)
        return new_children

    def process_blocknode(self, node, scope):
        assert hasattr(node, "children")
        node.children = self.merge_inline_sequences(node.children)
        for child in node.children:
            self.visit(child, scope)

    def visit_Document(self, node, scope):
        self.process_blocknode(node, scope)

    def visit_Paragraph(self, node, scope):
        self.process_blocknode(node, scope)

    def visit_ListItem(self, node, scope):
        self.process_blocknode(node, scope)

    def visit_TaskListItem(self, node, scope):
        self.process_blocknode(node, scope)

    def visit_BlockQuote(self, node, scope):
        self.process_blocknode(node, scope)

    def visit_Heading(self, node, scope):
        self.process_blocknode(node, scope)



class Node:
    def __init__(self):
        self.annotations = {}

    def get_children(self):
        return []

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class BlockNode(Node):
    def __init__(self, children):
        super().__init__()
        if children is None:
            self.children = []
        else:
            self.children = children if isinstance(children, list) else [children]

    def get_children(self):
        return self.children


# --- Text elements ---

class Text(Node):
    pass


class TextBlock(Text, BlockNode):
    pass


class LiteralText(Text):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __repr__(self):
        return f"LiteralText('{self.value}')"


class CodeSpan(Text):
    def __init__(self, code):
        self.code = code
        super().__init__()

    def __repr__(self):
        return f"CodeSpan(code='{self.code}')"


class Emphasis(TextBlock):
    def __repr__(self):
        return "Emphasis()"


class Strong(TextBlock):
    def __repr__(self):
        return "Strong()"


class StrikeThrough(TextBlock):
    def __repr__(self):
        return "StrikeThrough()"


class Link(TextBlock):
    def __init__(self, text, href, title=None):
        super().__init__(text)
        self.href = href
        self.title = title

    def __repr__(self):
        return f"Link(href='{self.href}', title='{self.title}')"


# --- Block elements ---


class Document(BlockNode):
    def __repr__(self):
        return "Document()"


class Paragraph(BlockNode):
    def __repr__(self):
        return "Paragraph()"


class Heading(BlockNode):
    def __init__(self, level, text):
        super().__init__(text)
        self.level = level

    def __repr__(self):
        return f"Heading(level={self.level})"


class BlockText(BlockNode):
    def __repr__(self):
        return "BlockText()"


class BlockQuote(BlockNode):
    def __repr__(self):
        return "BlockQuote()"


class BlockCode(BlockNode):
    def __init__(self, code=None, language=None, children=None):
        super().__init__(children)
        self.code = code
        self.language = language

    def __repr__(self):
        return f"BlockCode(language='{self.language}', code='{self.code}')"


# --- Image element ---


class Image(Node):
    def __init__(self, url, alt_text=None, title=None):
        super().__init__()
        self.url = url
        self.alt_text = alt_text
        self.title = title

    def get_children(self):
        return [self.alt_text]

    def __repr__(self):
        return (
            f"Image(url='{self.url}', title='{self.title}', alt_text='{self.alt_text}'"
        )


class ImageLink(Node):
    def __init__(self, img, href, title=None):
        super().__init__()
        self.img = img
        self.href = href
        self.title = title

    def get_children(self):
        return [self.img]

    def __repr__(self):
        return f"ImageLink(href='{self.href}', img='{self.img}', title='{self.title}')"


# --- Terminal elements ---


class Terminal(Node):
    pass


class TextTerminal(Text, Terminal):
    pass


class ThematicBreak(Terminal):
    def __repr__(self):
        return "ThematicBreak()"


class LineBreak(TextTerminal):
    def __repr__(self):
        return "LineBreak()"


class SoftBreak(TextTerminal):
    def __repr__(self):
        return "SoftBreak()"


class BlankLine(TextTerminal):
    def __repr__(self):
        return "BlankLine()"


# --- List elements ---


class List(Node):
    def __init__(self, elements, ordered=False):
        super().__init__()
        self.elements = elements
        self.ordered = ordered

    def get_children(self):
        return self.elements

    def __repr__(self):
        return f"List(ordered={self.ordered})"


class ListItem(BlockNode):
    def __init__(self, children):
        super().__init__(children)

    def __repr__(self):
        return "ListItem()"


class TaskListItem(BlockNode):
    def __init__(self, children, checked=False):
        super().__init__(children)
        self.checked = checked

    def __repr__(self):
        return f"TaskListItem(checked={self.checked})"


# --- Table elements ---


class Table(Node):
    def __init__(self, header=None, rows=None):
        super().__init__()
        self.header = header
        self.rows = rows

    def get_children(self):
        if self.header is None:
            if self.rows is None:
                children = []
            else:
                children = self.rows
        else:
            children = [self.header] + self.rows
        return children

    def __repr__(self):
        return "Table()"


class TableRow(Node):
    def __init__(self, row, is_header=False):
        super().__init__()
        self.row = row
        self.is_header = is_header

    def get_children(self):
        return self.row

    def __repr__(self):
        return f"TableRow(is_header={self.is_header})"


class TableCell(BlockNode):
    def __init__(self, cell, align=None):
        super().__init__(cell)
        self.align = align


class TableHeaderCell(TableCell):
    def __repr__(self):
        return f"TableHeaderCell(align={self.align})"


class TableRegularCell(TableCell):
    def __repr__(self):
        return f"TableRegularCell(align={self.align})"


class TableHeader(Node):
    def __init__(self, headers):
        super().__init__()
        self.headers = headers

    def get_children(self):
        return self.headers

    def __repr__(self):
        return "TableHeader()"



class ScopeStack:
    def __init__(self):
        self.stack = []

    def push(self, node):
        self.stack.append({"node": node})

    def pop(self, node):
        if not self.stack:
            raise RuntimeError("Scope underflow")
        top = self.stack.pop()
        if top["node"] != node:
            raise RuntimeError(
                "Scope mismatch: expected %s, got %s" % (top["node"], node)
            )
        return top

    def set(self, key, value):
        if not self.stack:
            raise RuntimeError("No active scope")
        self.stack[-1][key] = value

    def get(self, key, default=None):
        for frame in reversed(self.stack):
            if key in frame:
                return frame[key]
        return default

    def __repr__(self):
        lines = ["ScopeStack:"]
        for i, frame in enumerate(self.stack):
            node = frame.get("node")
            node_name = repr(node)
            other_keys = {k: v for k, v in frame.items() if k != "node"}
            if other_keys:
                lines.append(f"  [{i}] {node_name} {other_keys}")
            else:
                lines.append(f"  [{i}] {node_name}")
        return "\n".join(lines)



from mjml import mjml2html

from inkletter.md_to_mjml import parse_markdown_to_mjml


def parse_mjml_to_html(mjml_text):
    return mjml2html(mjml_text)


def parse_markdown_to_html(markdown_text):
    return parse_mjml_to_html(parse_markdown_to_mjml(markdown_text))



import mistune

from inkletter.ast import *
from inkletter.scope import ScopeStack
from inkletter.visitors.annotation import Annotation
from inkletter.visitors.merger import BlockTextMerger


class ASTRenderer(mistune.BaseRenderer):
    def render_token(self, token, state):
        func = self._get_method(token["type"])
        attrs = token.get("attrs")

        if "raw" in token:
            text = token["raw"]
        elif "children" in token:
            text = self.render_tokens(token["children"], state)
        else:
            text = None

        if attrs:
            if text is not None:
                return func(text, **attrs)
            return func(**attrs)
        else:
            if text is not None:
                return func(text)
            return func()

    def render_tokens(self, tokens, state):
        nodes = []
        for token in tokens:
            rv = self.render_token(token, state)
            if rv is None:
                continue
            if isinstance(rv, list):
                nodes.extend(rv)
            else:
                nodes.append(rv)
        return nodes

    def __call__(self, tokens, state):
        return Document(self.render_tokens(tokens, state))

    # --- Inline renderers ---

    def text(self, value):
        return LiteralText(value)

    def emphasis(self, text):
        return Emphasis(text)

    def strong(self, text):
        return Strong(text)

    def strikethrough(self, text):
        return StrikeThrough(text)

    def codespan(self, text):
        return CodeSpan(text)

    def link(self, text, url, title=None):
        # If the text contains an image, we create an ImageLink
        # We extract the image from the text to create a new ImageLink node
        # The rest of the text is ignored
        if img := next((t for t in text if isinstance(t, Image)), None):
            return ImageLink(img, url, title)
        else:
            return Link(text, url, title)

    # --- Block renderers ---

    def paragraph(self, text):
        return Paragraph(text)

    def heading(self, text, level):
        return Heading(level, text)

    def block_code(self, code, info=None):
        language = info.strip() if info else None
        return BlockCode(code, language)

    def block_text(self, text):
        return BlockText(text)

    def thematic_break(self):
        return ThematicBreak()

    def blank_line(self):
        return BlankLine()

    def linebreak(self):
        return LineBreak()

    def softbreak(self):
        return SoftBreak()

    def block_quote(self, text):
        return BlockQuote(text)

    def image(self, alt, url, title=None):
        if alt:
            # The tokenizer return always a list of a single text
            assert isinstance(alt, list), alt
            assert len(alt) == 1, alt
            alt = alt[0]
        return Image(url, alt, title)

    # --- List renderers ---

    def list(self, elements, ordered, **attrs):
        return List(elements, ordered=ordered)

    def list_item(self, element, **attrs):
        return ListItem(element)

    # --- Task renderers ---

    def task_list_item(self, text, checked=False):
        return TaskListItem(text, checked)

    # --- Table renderers ---

    def table(self, children):
        header = next(filter(lambda c: isinstance(c, TableHeader), children))
        rows = list(filter(lambda c: isinstance(c, TableRow), children))
        return Table(header=header, rows=rows)

    def table_body(self, text):
        return text

    def table_head(self, text):
        return TableHeader(text)

    def table_row(self, content):
        return TableRow(content)

    def table_cell(self, text, align=None, head=False):
        if head:
            return TableHeaderCell(text, align=align)
        else:
            return TableCell(text, align=align)


def parse_markdown_to_ast(markdown_text):
    renderer = ASTRenderer()
    markdown = mistune.create_markdown(
        renderer=renderer,
        plugins=[
            "mistune.plugins.formatting.strikethrough",
            "mistune.plugins.table.table",
            "mistune.plugins.table.table_in_list",
            "mistune.plugins.table.table_in_quote",
            "mistune.plugins.task_lists.task_lists",
        ],
    )
    ast = markdown(markdown_text)
    BlockTextMerger().visit(ast, scope=None)
    Annotation().visit(ast, scope=ScopeStack())
    return ast



from inkletter.visitors.generic import NodeVisitor


class TreeVisitor(NodeVisitor):
    def __init__(self):
        self.lines = []
        self._indent = 0

    def render(self, node):
        self.visit(node, scope=None)
        return "\n".join(self.lines)

    def write(self, text):
        self.lines.append("  " * self._indent + text)

    def generic_visit(self, node, scope=None):
        self.write(repr(node))
        self._indent += 1
        for child in node.get_children():
            self.generic_visit(child, scope)
        self._indent -= 1


def print_tree(doc):
    visitor = TreeVisitor()
    output = visitor.render(doc)
    print("\n" + output)



import pytest

from inkletter.codeblock import CodeBlock, CodeBlockResolver


def test_simple_function():
    cb = CodeBlock()
    cb.add_text("def foo():")
    cb.add_newline()
    cb.add_indent()
    cb.add_text("print('Hello')")
    cb.add_newline()
    cb.add_dedent()

    resolver = CodeBlockResolver()
    result = resolver.resolve(cb)

    expected = """\
def foo():
  print('Hello')"""
    assert result == expected


def test_nested_functions():
    cb = CodeBlock()
    cb.add_text("def outer():")
    cb.add_newline()
    cb.add_indent()
    cb.add_text("def inner():")
    cb.add_newline()
    cb.add_indent()
    cb.add_text("return 42")
    cb.add_newline()
    cb.add_dedent()
    cb.add_dedent()

    resolver = CodeBlockResolver()
    result = resolver.resolve(cb)

    expected = """\
def outer():
  def inner():
    return 42"""
    assert result == expected


def test_multiple_statements_same_indent():
    cb = CodeBlock()
    cb.add_text("x = 1")
    cb.add_newline()
    cb.add_text("y = 2")
    cb.add_newline()
    cb.add_text("z = x + y")
    cb.add_newline()

    resolver = CodeBlockResolver()
    result = resolver.resolve(cb)

    expected = """\
x = 1
y = 2
z = x + y"""
    assert result == expected


def test_indent_dedent_behavior():
    cb = CodeBlock()
    cb.add_text("if condition:")
    cb.add_newline()
    cb.add_indent()
    cb.add_text("do_something()")
    cb.add_newline()
    cb.add_dedent()
    cb.add_text("print('Done')")
    cb.add_newline()

    resolver = CodeBlockResolver()
    result = resolver.resolve(cb)

    expected = """\
if condition:
  do_something()
print('Done')"""
    assert result == expected


def test_empty_codeblock():
    cb = CodeBlock()

    resolver = CodeBlockResolver()
    result = resolver.resolve(cb)

    expected = ""  # Nothing generated
    assert result == expected


def test_consecutive_newlines():
    cb = CodeBlock()
    cb.add_text("first line")
    cb.add_newline()
    cb.add_newline()
    cb.add_text("second line")
    cb.add_newline()

    resolver = CodeBlockResolver()
    result = resolver.resolve(cb)

    expected = """\
first line

second line"""
    assert result == expected


def test_error_on_invalid_dedent():
    cb = CodeBlock()
    cb.add_dedent()

    resolver = CodeBlockResolver()

    with pytest.raises(ValueError, match="Cannot dedent below zero"):
        resolver.resolve(cb)
